@@ -258,6 +258,7 @@ menu_tree_add_menu_file_monitor (MenuTree   *tree,
     }
   else
     {
+      g_assert (handle == NULL);
       menu_verbose ("Failed to add file monitor to for %s: %s\n",
                     path, gnome_vfs_result_to_string (result));
     }
@@ -427,12 +428,12 @@ menu_tree_canonicalize_path (MenuTree *tree)
 static void
 menu_tree_force_recanonicalize (MenuTree *tree)
 {
+  menu_tree_remove_menu_file_monitors (tree);
+
   if (tree->canonical)
     {
       menu_tree_force_reload (tree);
 
-      menu_tree_remove_menu_file_monitors (tree);
-
       g_free (tree->canonical_path);
       tree->canonical_path = NULL;
 
@@ -559,11 +560,11 @@ find_path (MenuTreeDirectory *directory,
   char       *freeme;
   GSList     *tmp;
 
+  while (path[0] == '/') path++;
+
   if (path[0] == '\0')
     return directory;
 
-  while (path[0] == '/') path++;
-
   freeme = NULL;
   slash = strchr (path, '/');
   if (slash)
@@ -1116,6 +1117,9 @@ load_merge_file (MenuTree       *tree,
   merge_resolved_children (tree, where, to_merge);
 
   menu_layout_node_unref (to_merge);
+
+  if (freeme)
+    g_free (freeme);
 }
 
 static void
@@ -2225,6 +2229,16 @@ process_include_rules (MenuLayoutNode     *layout,
   return set;
 }
 
+static void
+mark_allocated_foreach (const char   *desktop_file_id,
+			DesktopEntry *desktop_entry,
+			GHashTable   *allocated)
+{
+  g_hash_table_insert (allocated,
+		       desktop_entry,
+		       desktop_entry_ref (desktop_entry));
+}
+
 static void
 entries_listify_foreach (const char        *desktop_file_id,
                          DesktopEntry      *desktop_entry,
@@ -2249,6 +2263,7 @@ process_layout (MenuTree          *tree,
   MenuLayoutNode     *layout_iter;
   MenuTreeDirectory  *directory;
   DesktopEntrySet    *entries;
+  DesktopEntrySet    *allocated_set;
   gboolean            deleted;
   gboolean            only_unallocated;
   GSList             *tmp;
@@ -2269,6 +2284,7 @@ process_layout (MenuTree          *tree,
   dir_dirs = menu_layout_node_menu_get_directory_dirs (layout);
 
   entries = desktop_entry_set_new ();
+  allocated_set = desktop_entry_set_new ();
 
   layout_iter = menu_layout_node_get_children (layout);
   while (layout_iter != NULL)
@@ -2307,6 +2323,7 @@ process_layout (MenuTree          *tree,
                 if (rule_set != NULL)
                   {
                     desktop_entry_set_union (entries, rule_set);
+                    desktop_entry_set_union (allocated_set, rule_set);
                     desktop_entry_set_unref (rule_set);
                   }
 
@@ -2392,6 +2409,15 @@ process_layout (MenuTree          *tree,
       layout_iter = menu_layout_node_get_next (layout_iter);
     }
 
+  directory->only_unallocated = only_unallocated;
+
+  if (!directory->only_unallocated)
+    desktop_entry_set_foreach (allocated_set,
+			       (DesktopEntrySetForeachFunc) mark_allocated_foreach,
+			       allocated);
+
+  desktop_entry_set_unref (allocated_set);
+
   if (directory->directory_entry)
     {
       if (desktop_entry_get_no_display (directory->directory_entry))
@@ -2416,8 +2442,6 @@ process_layout (MenuTree          *tree,
       return NULL;
     }
 
-  directory->only_unallocated = only_unallocated;
-
   directory->entries = NULL;
   desktop_entry_set_foreach (entries,
                              (DesktopEntrySetForeachFunc) entries_listify_foreach,
@@ -2469,22 +2493,6 @@ process_layout (MenuTree          *tree,
       tmp = next;
     }
 
-  if (!directory->only_unallocated)
-    {
-      /* Record the entries we just assigned to a node */
-      tmp = directory->entries;
-      while (tmp != NULL)
-        {
-          MenuTreeEntry *entry = tmp->data;
-
-          g_hash_table_insert (allocated,
-                               entry->desktop_entry,
-                               entry->desktop_entry);
-
-          tmp = tmp->next;
-        }
-    }
-
   g_assert (directory->name != NULL);
 
   return directory;
@@ -2569,7 +2577,8 @@ menu_tree_build_from_layout (MenuTree *tree)
 
   menu_verbose ("Building menu tree from layout\n");
 
-  allocated = g_hash_table_new (NULL, NULL);
+  allocated = g_hash_table_new_full (NULL, NULL, NULL,
+				     (GDestroyNotify) desktop_entry_unref);
 
   tree->root = process_layout (tree,
                                NULL,
